from pendulum import Date, now, parse

__all__ = (
    "left_strip",
    "right_strip",
    "split_lines",
    "starts_with",
    "parse_date",
    "today",
)

left_strip = str.lstrip
"""An alias of [`str.lstrip`][str.lstrip]."""
right_strip = str.rstrip
"""An alias of [`str.rstrip`][str.rstrip]."""

split_lines = str.splitlines
"""An alias of [`str.splitlines`][str.splitlines]."""
starts_with = str.startswith
"""An alias of [`str.startswith`][str.startswith]."""


def parse_date(string: str) -> Date:
    """Parses `YYYY-MM-DD` strings into dates.

    Arguments:
        string: The string to parse.

    Returns:
        The parsed date.
    """
    return parse(string).date()  # type: ignore  # pragma: no cover  # used in CLI


def today() -> Date:
    """Returns the current date.

    Returns:
        The current date.
    """
    return now().date()  # pragma: no cover  # not tested



from pathlib import Path

from typing_extensions import Literal

__all__ = (
    # constants
    "ROOT",
    "EMPTY",
    "SPACE",
    "NEW_LINE",
    "DOUBLE_NEW_LINE",
    "HASH",
    "DOT",
    "WRITE",
    # defaults
    "DEFAULT_QUIET",
    "DEFAULT_IGNORE_REQUIRED",
    "DEFAULT_NAME",
    "DEFAULT_ENCODING",
    "DEFAULT_ERRORS",
)

ROOT = Path(__file__).parent

EMPTY = str()
SPACE = " "

NEW_LINE = "\n"

DOUBLE_NEW_LINE = NEW_LINE + NEW_LINE

HASH = "#"

DOT = "."

WRITE: Literal["w"] = "w"

DEFAULT_QUIET = True

DEFAULT_IGNORE_REQUIRED = False

DEFAULT_NAME = "changelogging"

DEFAULT_ENCODING = "utf-8"
DEFAULT_ERRORS = "strict"



from pathlib import Path
from subprocess import call
from typing import Iterable, Sequence

from iters.utils import unary_tuple
from typing_aliases import DynamicTuple

from changelogging.constants import DEFAULT_QUIET

__all__ = ("remove_command", "remove_paths")

GIT = "git"
REMOVE = "rm"
QUIET = "-q"
FORCE = "-f"


def resolve_path(path: Path) -> str:
    return path.resolve().as_posix()


def remove_command(iterable: Iterable[Path], quiet: bool = DEFAULT_QUIET) -> Sequence[str]:
    command: DynamicTuple[str] = (GIT, REMOVE, FORCE)

    if quiet:
        command += unary_tuple(QUIET)

    command += tuple(map(resolve_path, iterable))

    return command


def remove_paths(iterable: Iterable[Path], quiet: bool = DEFAULT_QUIET) -> None:
    call(remove_command(iterable, quiet=quiet))



from collections import defaultdict as default_dict
from pathlib import Path
from textwrap import wrap
from typing import Dict, Iterable, Iterator, List, Mapping, Tuple, TypeVar

from attrs import define, field
from funcs.unpacking import unpack_binary
from iters.iters import iter, wrap_iter
from pendulum import Date
from typing_aliases import Binary, Unary

from changelogging.config import Config
from changelogging.constants import (
    DEFAULT_ENCODING,
    DEFAULT_ERRORS,
    DOT,
    DOUBLE_NEW_LINE,
    EMPTY,
    HASH,
    NEW_LINE,
    SPACE,
    WRITE,
)
from changelogging.fragments import Fragment, FragmentType, Issue
from changelogging.utils import left_strip, split_lines, today

__all__ = ("Builder",)

NO_SIGNIFICANT_CHANGES = "No significant changes."

FT = TypeVar("FT", bound=FragmentType)
IT = TypeVar("IT", bound=Issue)

Fragments = Iterable[Fragment[FT, IT]]
Sections = Mapping[FT, Fragments[FT, IT]]


def get_path(item: Tuple[Path, FragmentType]) -> Path:  # pragma: no cover  # `main` only
    path, _ = item

    return path


@define()
class Builder:
    """Represents changelog builders."""

    config: Config = field()
    """The config of the builder."""

    date: Date = field(factory=today)
    """The date to use in builds."""

    def build_title(self) -> str:
        """Builds the main title.

        Returns:
            The main title.
        """
        config = self.config

        return self.heading(config.title_level) + config.title_format.format(
            name=config.name, version=config.version, url=config.url, date=self.date
        )

    def build_section_title(self, type: FragmentType) -> str:
        """Builds the section title of the particular `type`.

        Arguments:
            type: The fragment type to build the title for.

        Returns:
            The section title.
        """
        config = self.config

        return self.heading(config.section_level) + type.title

    def build_issue(self, issue: Issue) -> str:
        """Builds the `issue` given.

        Arguments:
            issue: The issue to build.

        Returns:
            The issue built.
        """
        config = self.config

        return config.issue_format.format(issue=issue.value, url=config.url)

    def build_fragment(self, fragment: Fragment[FragmentType, Issue]) -> str:
        """Builds the `fragment` given.

        Arguments:
            fragment: The fragment to build.

        Returns:
            The built fragment.
        """
        config = self.config

        content = config.fragment_format.format(
            content=fragment.content.strip(), issue=self.build_issue(fragment.issue)
        )

        return (
            self.indent_wrap_lines(content, config.bullet, config.wrap_size)
            if config.wrap
            else self.indent_lines(content, config.bullet)
        )

    def build_fragments(self, fragments: Fragments[FragmentType, Issue]) -> str:
        """Builds several `fragments` given.

        Arguments:
            fragments: The fragments to build.

        Returns:
            The fragments built.
        """
        return iter(fragments).map(self.build_fragment).join(NEW_LINE)

    def collect_sections(self, fragments: Fragments[FT, IT]) -> Sections[FT, IT]:
        """Collects `fragments` into sections.

        Arguments:
            fragments: The fragments to collect.

        Returns:
            The collected sections.
        """
        sections: Dict[FT, List[Fragment[FT, IT]]] = default_dict(list)

        for fragment in fragments:
            sections[fragment.type].append(fragment)

        for fragments in sections.values():
            fragments.sort()

        return sections

    @wrap_iter
    def build_generate(self, sections: Sections[FragmentType, Issue]) -> Iterator[str]:
        """Builds `sections`, returning an iterator.

        Arguments:
            sections: The sections to build.

        Returns:
            The iterator over the build result.
        """
        config = self.config

        yield self.build_title()

        empty = True

        for type in config.display.into_types(config.types):
            if type in sections:
                empty = False

                fragments = sections[type]

                yield self.build_section_title(type)
                yield self.build_fragments(fragments)

        if empty:  # pragma: no cover  # not tested
            yield NO_SIGNIFICANT_CHANGES

    def fetch_fragment(
        self,
        path: Path,
        type: FragmentType,
        encoding: str = DEFAULT_ENCODING,
        errors: str = DEFAULT_ERRORS,
    ) -> Fragment[FragmentType, Issue]:
        """Fetches a fragment of `type` from `path`.

        Arguments:
            path: The path to fetch the fragment from.
            type: The type of the fragment to fetch.
            encoding: The encoding to use.
            errors: The error handling strategy to use.

        Returns:
            The fetched fragment.
        """
        issue = self.get_issue(path)
        content = path.read_text(encoding, errors)

        return Fragment(type, content, issue)

    def fetching_fragment(
        self, encoding: str = DEFAULT_ENCODING, errors: str = DEFAULT_ERRORS
    ) -> Binary[Path, FragmentType, Fragment[FragmentType, Issue]]:
        def fetch_fragment(path: Path, type: FragmentType) -> Fragment[FragmentType, Issue]:
            return self.fetch_fragment(path, type, encoding, errors)

        return fetch_fragment

    @wrap_iter
    def collect_fragments(
        self, encoding: str = DEFAULT_ENCODING, errors: str = DEFAULT_ERRORS
    ) -> Fragments[FragmentType, Issue]:
        """Collects fragments from the changes directory specified in the config.

        Arguments:
            encoding: The encoding to use.
            errors: The error handling strategy to use.

        Returns:
            The iterator over the fragments found.
        """
        return (
            self.collect_paths_types()
            .map(unpack_binary(self.fetching_fragment(encoding, errors)))
            .unwrap()
        )

    @wrap_iter
    def collect_paths_types(self) -> Iterator[Tuple[Path, FragmentType]]:
        config = self.config
        directory = config.directory
        types = config.types

        # TODO: perhaps rewrite this using iterators?

        for path in directory.iterdir():
            if path.is_file():
                for suffix in path.suffixes:
                    if types.has_suffix(suffix):
                        yield (path, types.get_suffix(suffix))

    @wrap_iter
    def collect_paths(self) -> Iterator[Path]:
        """Collect paths to fragments.

        Returns:
            The iterator over paths found.
        """
        return self.collect_paths_types().map(get_path).unwrap()  # pragma: no cover  # `main` only

    @classmethod
    def get_issue(cls, path: Path) -> Issue:
        return Issue(int(cls.name_no_suffixes(path)))

    def build(self) -> str:
        """Builds the changelog.

        Returns:
            The build result.
        """
        return self.build_generate(self.collect_sections(self.collect_fragments().unwrap())).join(
            DOUBLE_NEW_LINE
        )

    def write(self, encoding: str = DEFAULT_ENCODING, errors: str = DEFAULT_ERRORS) -> None:
        """Builds the changelog and writes it to the output file.

        Arguments:
            encoding: The encoding to use.
            errors: The error handling strategy to use.
        """
        config = self.config

        output = config.output
        start_string = config.start_string

        content = self.build()

        current = output.read_text(encoding, errors) if output.exists() else EMPTY

        before, start, after = current.partition(start_string)

        with output.open(WRITE, encoding=encoding, errors=errors) as file:
            if not start:  # pragma: no cover  # not tested
                file.write(content + NEW_LINE)

                if current.strip():
                    file.write(NEW_LINE + left_strip(current))

            else:
                file.write(before)
                file.write(start)
                file.write(DOUBLE_NEW_LINE + content + NEW_LINE)

                if after.strip():
                    file.write(NEW_LINE + left_strip(after))

    @staticmethod
    def heading(level: int) -> str:
        return HASH * level + SPACE

    @staticmethod
    def indents(bullet: str) -> Tuple[str, str]:
        space = SPACE
        return (bullet + space, space + space)

    @staticmethod
    def name_no_suffixes(path: Path) -> str:
        name, _, _ = path.name.partition(DOT)
        return name

    @classmethod
    @wrap_iter
    def generate_indent_lines(cls, string: str, bullet: str) -> Iterator[str]:
        if not string:  # pragma: no cover  # not tested
            return

        initial, subsequent = cls.indents(bullet)

        head, *tail = split_lines(string)

        yield (initial + head if head.strip() else head)

        for item in tail:
            yield (subsequent + item if item.strip() else item)

    @classmethod
    def indent_lines(cls, string: str, bullet: str) -> str:
        return cls.generate_indent_lines(string, bullet).join(NEW_LINE)

    @staticmethod
    def wrapping_line(size: int) -> Unary[str, str]:
        def wrap_line(line: str) -> str:
            return iter(wrap(line, size, break_long_words=False)).join(NEW_LINE)

        return wrap_line

    @classmethod
    def indent_wrap_lines(cls, string: str, bullet: str, size: int) -> str:
        size -= iter(cls.indents(bullet)).map(len).max().unwrap()

        return cls.indent_lines(
            iter(split_lines(string)).map(cls.wrapping_line(size)).join(NEW_LINE), bullet
        )



from pathlib import Path
from typing import Any, Iterable, Optional, Type, TypeVar, overload

from attrs import define
from iters.iters import iter
from iters.utils import contains_only_item, empty
from toml import loads as load_string
from typing_aliases import IntoPath, StringDict, Unary
from typing_extensions import Self
from versions.functions import parse_version
from versions.version import Version
from wraps.option import Option, Some
from wraps.wraps import wrap_optional
from yarl import URL

from changelogging.constants import (
    DEFAULT_ENCODING,
    DEFAULT_ERRORS,
    DEFAULT_IGNORE_REQUIRED,
    EMPTY,
    ROOT,
)
from changelogging.fragments import AnyFragmentTypes, Display, FragmentType

__all__ = ("Config", "ConfigData", "AnyConfigData")

CHANGELOGGING = "changelogging.toml"
PYPROJECT = "pyproject.toml"

SEARCH = (CHANGELOGGING, PYPROJECT)

DEFAULT_PATH = ROOT / CHANGELOGGING

CONFIG_NOT_FOUND = "can not find config in {}"
EXPECTED_FILE_OR_DIRECTORY = "expected either a file or a directory"

EXPECTED = "expected `{}`"
expected = EXPECTED.format

EXPECTED_ONE_CHARACTER_BULLET = "expected one character bullet"

EXPECTED_CHANGELOGGING = expected("changelogging")
EXPECTED_CHANGELOGGING_NAME = expected("changelogging.name")
EXPECTED_CHANGELOGGING_VERSION = expected("changelogging.version")
EXPECTED_CHANGELOGGING_URL = expected("changelogging.url")
EXPECTED_CHANGELOGGING_DIRECTORY = expected("changelogging.directory")
EXPECTED_CHANGELOGGING_OUTPUT = expected("changelogging.output")
EXPECTED_CHANGELOGGING_TITLE_LEVEL = expected("changelogging.title_level")
EXPECTED_CHANGELOGGING_SECTION_LEVEL = expected("changelogging.section_level")
EXPECTED_CHANGELOGGING_BULLET = expected("changelogging.bullet")
EXPECTED_CHANGELOGGING_WRAP = expected("changelogging.wrap")
EXPECTED_CHANGELOGGING_WRAP_SIZE = expected("changelogging.wrap_size")
EXPECTED_CHANGELOGGING_START_STRING = expected("changelogging.start_string")
EXPECTED_CHANGELOGGING_TITLE_FORMAT = expected("changelogging.title_format")
EXPECTED_CHANGELOGGING_ISSUE_FORMAT = expected("changelogging.issue_format")
EXPECTED_CHANGELOGGING_FRAGMENT_FORMAT = expected("changelogging.fragment_format")
EXPECTED_CHANGELOGGING_DISPLAY = expected("changelogging.display")
EXPECTED_CHANGELOGGING_TYPES_TYPE_NAME = expected("changelogging.types.type.name")
EXPECTED_CHANGELOGGING_TYPES_TYPE_TITLE = expected("changelogging.types.type.title")


def config_not_found(path: Path) -> FileNotFoundError:
    return FileNotFoundError(CONFIG_NOT_FOUND.format(path.resolve().as_posix()))


def expected_file_or_directory() -> ValueError:  # pragma: no cover
    return ValueError(EXPECTED_FILE_OR_DIRECTORY)


T = TypeVar("T")


class ConfigData(StringDict[T]):
    """Dictionaries that support attribute access."""

    def __getattr__(self, name: str) -> Option[T]:
        return wrap_optional(self.get(name))

    def copy(self) -> Self:
        return type(self)(self)


AnyConfigData = ConfigData[Any]


FT = TypeVar("FT", bound=FragmentType)


@overload
def mapping_to_type(mapping: ConfigData[str]) -> FragmentType:
    ...


@overload
def mapping_to_type(mapping: ConfigData[str], fragment_type: Type[FT]) -> FT:
    ...


def mapping_to_type(mapping: ConfigData[str], fragment_type: Type[Any] = FragmentType) -> Any:
    return fragment_type(
        mapping.name.expect(EXPECTED_CHANGELOGGING_TYPES_TYPE_NAME),
        mapping.title.expect(EXPECTED_CHANGELOGGING_TYPES_TYPE_TITLE),
    )


def validate_bullet(bullet: str) -> str:
    return Some(bullet).filter(contains_only_item).expect(EXPECTED_ONE_CHARACTER_BULLET)


@define()
class Config:
    name: str
    """The name of the project."""
    version: Version
    """The version of the project."""
    url: URL
    """The URL of the project."""
    directory: Path
    """The `changes` directory."""
    output: Path
    """The output path."""
    title_level: int
    """The title level to use."""
    section_level: int
    """The section title level to use."""
    bullet: str
    """The bullet to use."""
    wrap: bool
    """Whether to wrap lines."""
    wrap_size: int
    """The wrap size to use."""
    start_string: str
    """The start string to look for."""
    title_format: str
    """The format of the title."""
    issue_format: str
    """The format of the issue."""
    fragment_format: str
    """The format of the fragment."""
    display: Display
    """The display ordering of fragment types."""
    types: AnyFragmentTypes
    """The fragment types to use."""

    # dynamic code ahead...

    @classmethod
    def from_string(cls, string: str, source: Optional[Path] = None) -> Self:
        """Parses a [`Config`][changelogging.config.Config] from `string`.

        Arguments:
            string: The string to parse.
            source: The source of where the config came from.

        Returns:
            The newly parsed [`Config`][changelogging.config.Config].
        """
        return cls.from_data(cls.parse(string), source)

    @classmethod
    def from_file_path(
        cls, path: IntoPath, encoding: str = DEFAULT_ENCODING, errors: str = DEFAULT_ERRORS
    ) -> Self:
        """Parses a [`Config`][changelogging.config.Config] from file `path`.

        Arguments:
            path: The path to the config.
            encoding: The encoding to use.
            errors: The error handling strategy to use.

        Returns:
            The newly parsed [`Config`][changelogging.config.Config] instance.
        """
        path = Path(path)

        return cls.from_string(path.read_text(encoding, errors), path)

    @classmethod
    def from_path(
        cls,
        path: IntoPath,
        search: Iterable[IntoPath] = SEARCH,
        encoding: str = DEFAULT_ENCODING,
        errors: str = DEFAULT_ERRORS,
    ) -> Self:
        """Parses a [`Config`][changelogging.config.Config] from `path`.

        If `path` is a directory, this function searches for files in `search` inside of it.

        Arguments:
            path: The path to the config.
            search: The paths to search for.
            encoding: The encoding to use.
            errors: The error handling strategy to use.

        Returns:
            The newly parsed [`Config`][changelogging.config.Config].
        """
        path = Path(path)

        if not path.exists():
            raise config_not_found(path)

        if path.is_dir():
            for part in search:
                try_path = path / part

                if try_path.exists():
                    return cls.from_file_path(try_path, encoding, errors)

            raise config_not_found(path)

        if path.is_file():
            return cls.from_file_path(path, encoding, errors)

        raise expected_file_or_directory()  # pragma: no cover

    @staticmethod
    def parse(string: str) -> AnyConfigData:
        return load_string(string, AnyConfigData)

    @classmethod
    def from_data(cls, config_data: AnyConfigData, source: Optional[Path] = None) -> Self:
        """Creates a [`Config`][changelogging.config.Config]
        from [`ConfigData`][changelogging.config.ConfigData].

        Arguments:
            config_data: The config data to use.

        Returns:
            The newly created [`Config`][changelogging.config.Config] instance.
        """
        config_data = config_data.tool.unwrap_or(config_data)
        config = config_data.changelogging.expect(EXPECTED_CHANGELOGGING)

        types = (
            iter(config.types.unwrap_or_else(empty))
            .map(mapping_to_type)
            .collect(AnyFragmentTypes.from_iterable)
        )

        default_config = DEFAULT_CONFIG

        return cls(
            # `name`, `version` and `url` are always required
            name=config.name.expect(EXPECTED_CHANGELOGGING_NAME),
            version=config.version.map(parse_version).expect(EXPECTED_CHANGELOGGING_VERSION),
            # map to `URL` and `Path` to simplify interaction
            url=config.url.map(URL).expect(EXPECTED_CHANGELOGGING_URL),
            directory=config.directory.map_or(default_config.directory, source_path(source)),
            output=config.output.map_or(default_config.output, source_path(source)),
            # merely return defaults if needed
            title_level=config.title_level.unwrap_or(default_config.title_level),
            section_level=config.section_level.unwrap_or(default_config.section_level),
            bullet=validate_bullet(config.bullet.unwrap_or(default_config.bullet)),
            wrap=config.wrap.unwrap_or(default_config.wrap),
            wrap_size=config.wrap_size.unwrap_or(default_config.wrap_size),
            start_string=config.start_string.unwrap_or(default_config.start_string),
            title_format=config.title_format.unwrap_or(default_config.title_format),
            issue_format=config.issue_format.unwrap_or(default_config.issue_format),
            fragment_format=config.fragment_format.unwrap_or(default_config.fragment_format),
            # map to `Display` type
            display=config.display.map_or(default_config.display, Display.from_iterable),
            # merge user-defined `types` with already existing default ones
            types=default_config.types.merge_with(types),
        )

    @classmethod
    def unsafe_from_string(
        cls,
        string: str,
        source: Optional[Path] = None,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> Self:
        return cls.unsafe_from_data(cls.parse(string), source, ignore_required=ignore_required)

    @classmethod
    def unsafe_from_file_path(
        cls,
        path: IntoPath,
        encoding: str = DEFAULT_ENCODING,
        errors: str = DEFAULT_ERRORS,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> Self:
        path = Path(path)

        return cls.unsafe_from_string(
            path.read_text(encoding, errors), path, ignore_required=ignore_required
        )

    @classmethod
    def unsafe_from_path(
        cls,
        path: IntoPath,
        search: Iterable[IntoPath] = SEARCH,
        encoding: str = DEFAULT_ENCODING,
        errors: str = DEFAULT_ERRORS,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> Self:
        path = Path(path)

        if not path.exists():  # pragma: no cover
            raise config_not_found(path)

        if path.is_dir():  # pragma: no cover
            for part in search:
                try_path = path / part

                if try_path.exists():
                    return cls.unsafe_from_file_path(
                        try_path, encoding, errors, ignore_required=ignore_required
                    )

            raise config_not_found(path)

        if path.is_file():
            return cls.unsafe_from_file_path(
                path, encoding, errors, ignore_required=ignore_required
            )

        raise expected_file_or_directory()  # pragma: no cover

    @classmethod
    def unsafe_from_data(
        cls,
        config_dict: AnyConfigData,
        source: Optional[Path] = None,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> Self:
        config_dict = config_dict.tool.unwrap_or(config_dict)
        config = config_dict.changelogging.expect(EXPECTED_CHANGELOGGING)

        types = (
            iter(config.types.unwrap_or_else(empty))
            .map(mapping_to_type)
            .collect(AnyFragmentTypes.from_iterable)
        )

        name_option = config.name

        name = (
            name_option.unwrap_or(EMPTY)
            if ignore_required
            else name_option.expect(EXPECTED_CHANGELOGGING_NAME)
        )

        version_option = config.version.map(parse_version)

        version = (
            version_option.unwrap_or_else(Version)
            if ignore_required
            else version_option.expect(EXPECTED_CHANGELOGGING_VERSION)
        )

        url_option = config.url.map(URL)

        url = (
            url_option.unwrap_or_else(URL)
            if ignore_required
            else url_option.expect(EXPECTED_CHANGELOGGING_URL)
        )

        return cls(
            name=name,
            version=version,
            url=url,
            directory=config.directory.map(source_path(source)).expect(
                EXPECTED_CHANGELOGGING_DIRECTORY
            ),
            output=config.output.map(source_path(source)).expect(EXPECTED_CHANGELOGGING_OUTPUT),
            title_level=config.title_level.expect(EXPECTED_CHANGELOGGING_TITLE_LEVEL),
            section_level=config.section_level.expect(EXPECTED_CHANGELOGGING_SECTION_LEVEL),
            bullet=validate_bullet(config.bullet.expect(EXPECTED_CHANGELOGGING_BULLET)),
            wrap=config.wrap.expect(EXPECTED_CHANGELOGGING_WRAP),
            wrap_size=config.wrap_size.expect(EXPECTED_CHANGELOGGING_WRAP_SIZE),
            start_string=config.start_string.expect(EXPECTED_CHANGELOGGING_START_STRING),
            title_format=config.title_format.expect(EXPECTED_CHANGELOGGING_TITLE_FORMAT),
            issue_format=config.issue_format.expect(EXPECTED_CHANGELOGGING_ISSUE_FORMAT),
            fragment_format=config.fragment_format.expect(EXPECTED_CHANGELOGGING_FRAGMENT_FORMAT),
            display=config.display.map(Display.from_iterable).expect(
                EXPECTED_CHANGELOGGING_DISPLAY
            ),
            types=types,
        )


def source_path(source: Optional[Path] = None) -> Unary[str, Path]:
    def format_path(string: str) -> Path:
        if source is None:
            return Path(string)

        return Path(string.format(here=source.parent))

    return format_path


DEFAULT_CONFIG = Config.unsafe_from_path(DEFAULT_PATH, ignore_required=True)



"""Building changelogs from fragments."""

__description__ = "Building changelogs from fragments."
__url__ = "https://github.com/nekitdev/changelogging"

__title__ = "changelogging"
__author__ = "nekitdev"
__license__ = "MIT"
__version__ = "1.4.1"

from changelogging.build import Builder
from changelogging.config import Config
from changelogging.fragments import Display, Fragment, FragmentType, FragmentTypes, Issue

__all__ = ("Builder", "Config", "Display", "Fragment", "FragmentType", "FragmentTypes", "Issue")



from pathlib import Path
from typing import Optional

import click
from iters.iters import iter
from pendulum import Date
from wraps.option import Option
from wraps.wraps import wrap_optional

from changelogging import __version__ as version
from changelogging.build import Builder
from changelogging.config import Config
from changelogging.constants import DEFAULT_NAME, DEFAULT_QUIET, HASH, NEW_LINE
from changelogging.git import remove_paths
from changelogging.utils import parse_date, right_strip, split_lines, starts_with, today

__all__ = ("changelogging", "build", "create")


def get_config(string: Option[str]) -> Config:
    return Config.from_path(string.map_or_else(Path, Path))


def get_date(string: Option[str]) -> Date:
    return string.map_or_else(today, parse_date)


@click.group(name=DEFAULT_NAME)
@click.help_option("--help", "-h")
@click.version_option(version, "--version", "-V")
def changelogging() -> None:
    pass


CONFIG_PATH = "config_path"
DATE_STRING = "date_string"

REMOVING = "removing `{}`"
removing = REMOVING.format


@changelogging.command()
@click.help_option("--help", "-h")
@click.option("--config", "-c", CONFIG_PATH, default=None)
@click.option("--date", "-d", DATE_STRING, default=None)
@click.option("--quiet", "-q", is_flag=True, default=DEFAULT_QUIET)
@click.option("--draft", "-D", is_flag=True, default=False)
@click.option("--remove/--no-remove", "-r/-n", default=False)
def build(
    config_path: Optional[str], date_string: Optional[str], quiet: bool, draft: bool, remove: bool
) -> None:
    date = get_date(wrap_optional(date_string))

    config = get_config(wrap_optional(config_path))

    builder = Builder(config, date)

    if draft:
        click.echo(builder.build())

    else:
        if remove:
            remove_paths(builder.collect_paths(), quiet=quiet)

        builder.write()


ABORTED = "Creation aborted."
CREATED = "Created the `{}` fragment."
PLACEHOLDER = "Add the content here."
EDIT = """
# Please enter the fragment content.
# Lines starting with `#` will be ignored.
# Close the file without saving to abort.
"""

NAME = "name"


def is_comment(line: str) -> bool:
    return starts_with(line, HASH)


def is_content(line: str) -> bool:
    return not is_comment(line)


@changelogging.command()
@click.help_option("--help", "-h")
@click.option("--config", "-c", CONFIG_PATH, default=None)
@click.option("--edit/--no-edit", "-e/-n", default=True)
@click.argument(NAME)
def create(config_path: Optional[str], edit: bool, name: str) -> None:
    config = get_config(wrap_optional(config_path))

    if edit:
        string = click.edit(EDIT)

        if string is None:
            click.echo(ABORTED)

            return

        string = iter(split_lines(string)).filter(is_content).map(right_strip).join(NEW_LINE)

    else:
        string = PLACEHOLDER

    path = config.directory / name

    path.write_text(right_strip(string) + NEW_LINE)

    click.echo(CREATED.format(name))



from typing import Any, Generic, Iterable, Iterator, TypeVar

from attrs import field, frozen
from iters.mappings import merge
from typing_aliases import DynamicTuple, StringDict
from typing_extensions import Self

__all__ = ("Display", "Fragment", "FragmentType", "FragmentTypes", "AnyFragmentTypes", "Issue")

SUFFIX = ".{}"
suffix = SUFFIX.format


@frozen()
class FragmentType:
    """Represents fragment types."""

    name: str
    """The name of the type."""
    title: str
    """The title of the type."""

    @property
    def suffix(self) -> str:
        """The suffix of the type."""
        return suffix(self.name)


FT = TypeVar("FT", bound=FragmentType)

FragmentTypeTuple = DynamicTuple[FT]
FragmentTypeDict = StringDict[FT]


@frozen()
class FragmentTypes(Generic[FT]):
    """Represents collections of fragment types."""

    types: FragmentTypeTuple[FT] = ()
    """The contained fragment types."""

    @classmethod
    def from_types(cls, *types: FT) -> Self:
        """Creates [`FragmentTypes`][changelogging.fragments.FragmentTypes] from `types`.

        Arguments:
            *types: Fragment types to collect.

        Returns:
            A newly created [`FragmentTypes`][changelogging.fragments.FragmentTypes] instance.
        """
        return cls(types)

    @classmethod
    def from_iterable(cls, iterable: Iterable[FT]) -> Self:
        """Creates [`FragmentTypes`][changelogging.fragments.FragmentTypes]
        from an `iterable` of types.

        Arguments:
            iterable: The iterable to collect types from.

        Returns:
            A newly created [`FragmentTypes`][changelogging.fragments.FragmentTypes] instance.
        """
        return cls(tuple(iterable))

    def merge_with(self, other: Self) -> Self:
        """Merges `self` and `other` into one instance.

        Arguments:
            other: The types to merge with `self`.

        Returns:
            The merged [`FragmentTypes`][changelogging.fragments.FragmentTypes] instance.
        """
        merged = merge(self.name_to_type, other.name_to_type)

        return self.from_iterable(merged.values())

    @property
    def name_to_type(self) -> FragmentTypeDict[FT]:
        """The `name -> type` mapping."""
        return {type.name: type for type in self.types}

    @property
    def suffix_to_type(self) -> FragmentTypeDict[FT]:
        """The `suffix -> type` mapping."""
        return {type.suffix: type for type in self.types}

    def has_name(self, name: str) -> bool:
        """Checks if the `name` is in [`FragmentTypes`][changelogging.fragments.FragmentTypes].

        Arguments:
            name: The name to check.

        Returns:
            Whether the name is present.
        """
        return name in self.name_to_type

    def get_name(self, name: str) -> FT:
        """Finds a [`FragmentType`][changelogging.fragments.FragmentType] by `name`.

        Arguments:
            name: The name to lookup.

        Raises:
            KeyError: Fragment type was not found.

        Returns:
            The fragment type found.
        """
        return self.name_to_type[name]

    def has_suffix(self, suffix: str) -> bool:
        """Checks if the `suffix` is in [`FragmentTypes`][changelogging.fragments.FragmentTypes].

        Arguments:
            suffix: The suffix to check.

        Returns:
            Whether the suffix is present.
        """
        return suffix in self.suffix_to_type

    def get_suffix(self, suffix: str) -> FT:
        """Finds a [`FragmentType`][changelogging.fragments.FragmentType] by `suffix`.

        Arguments:
            suffix: The suffix to lookup.

        Raises:
            KeyError: Fragment type was not found.

        Returns:
            The fragment type found.
        """
        return self.suffix_to_type[suffix]


AnyFragmentTypes = FragmentTypes[Any]

Names = DynamicTuple[str]

TYPE_NOT_FOUND = "can not find `{}` type"


def type_not_found(name: str) -> LookupError:
    return LookupError(TYPE_NOT_FOUND.format(name))


@frozen()
class Display:
    """Represents ordering of fragments' display."""

    names: Names = ()
    """The names of the fragment types."""

    @classmethod
    def from_names(cls, *names: str) -> Self:
        """Creates a [`Display`][changelogging.fragments.Display] from `names`.

        Arguments:
            *names: The names to use.

        Returns:
            A newly created [`Display`][changelogging.fragments.Display] instance.
        """
        return cls(names)

    @classmethod
    def from_iterable(cls, iterable: Iterable[str]) -> Self:
        """Creates a [`Display`][changelogging.fragments.Display] from an `iterable` of names.

        Arguments:
            iterable: The iterable to collect names from.

        Returns:
            A newly created [`Display`][changelogging.fragments.Display] instance.
        """
        return cls(tuple(iterable))

    def into_types(self, types: FragmentTypes[FT]) -> Iterator[FT]:
        """Convert names into respective fragment types according to `types`.

        Arguments:
            types: The fragment types to search in.

        Raises:
            LookupError: Fragment type was not found.

        Returns:
            The iterator over fragment types found.
        """
        for name in self.names:
            if types.has_name(name):
                yield types.get_name(name)

            else:
                raise type_not_found(name)


@frozen(order=True)
class Issue:
    """Represents issues."""

    value: int
    """The value of the issue."""


IT = TypeVar("IT", bound=Issue)


@frozen(order=True)
class Fragment(Generic[FT, IT]):
    """Represents fragments."""

    type: FT = field(order=False)
    """The type of the fragment."""
    content: str = field(order=False)
    """The content of the fragment."""
    issue: IT = field(order=True)
    """The issue related to the fragment."""
